# TODO: add here environment configuration for local testing.

# mainnet
# https://rpc.mainnet.near.org
# testnet
# https://rpc.testnet.near.org
# betanet (may be unstable)
# https://rpc.betanet.near.org
# localnet
# http://localhost: 3030

# NODE_URL = "https://rpc.testnet.near.org"

NODE_URL = "https://rpc.ci-testnet.near.org"
TEST_ACCOUNT = "test.near"
TEST_KEY_PAIR = "ed25519:2wyRcSwSuHtRVmkMCGjPwnzZmQLeXLzLLyED1NDMt4BjnKgQL6tF85yBx6Jr26D2dUNeC716RBoTxntVHsegogYw"
